import { useMemo } from 'react';
import * as THREE from 'three';
import * as turf from '@turf/turf';

export function useWallClassification(walls: Walls) {
    // Find all minimal rooms from the given walls
    const findRooms = () => {
        if (walls.length < 3) return [];

        // Build a graph of point connections
        const graph = new Map<string, string[]>();
        const pointMap = new Map<string, Point>();

        walls.forEach(wall => {
            const [p1, p2] = wall.points;
            pointMap.set(p1.pointUuid, p1);
            pointMap.set(p2.pointUuid, p2);

            // Add connection from p1 to p2
            if (!graph.has(p1.pointUuid)) graph.set(p1.pointUuid, []);
            graph.get(p1.pointUuid)?.push(p2.pointUuid);

            // Add connection from p2 to p1
            if (!graph.has(p2.pointUuid)) graph.set(p2.pointUuid, []);
            graph.get(p2.pointUuid)?.push(p1.pointUuid);
        });

        // Find all minimal cycles (rooms) in the graph
        const allCycles: string[][] = [];
        const visited = new Set<string>();

        const findCycles = (startNode: string, currentNode: string, path: string[], depth = 0) => {
            if (depth > 20) return; // Prevent infinite recursion

            visited.add(currentNode);
            path.push(currentNode);

            const neighbors = graph.get(currentNode) || [];

            for (const neighbor of neighbors) {
                if (path.length > 2 && neighbor === startNode) {
                    // Found a cycle that returns to start
                    const cycle = [...path, neighbor];

                    // Check if this is a new unique cycle
                    if (!cycleExists(allCycles, cycle)) {
                        allCycles.push(cycle);
                    }
                    continue;
                }

                if (!path.includes(neighbor)) {
                    findCycles(startNode, neighbor, [...path], depth + 1);
                }
            }
        };

        // Start cycle detection from each node
        for (const [pointUuid] of graph) {
            findCycles(pointUuid, pointUuid, []);
        }

        // Convert cycles to Point arrays and validate them
        const potentialRooms = allCycles
            .map(cycle => cycle.map(uuid => pointMap.get(uuid)!))
            .filter(room => isValidRoom(room));

        // Filter out duplicate rooms (same set of points in different orders)
        const uniqueRooms = removeDuplicateRooms(potentialRooms);

        // Remove rooms that fully contain other rooms (i.e., outer enclosing rooms)
        const turfRooms = uniqueRooms.map(room => ({
            points: room,
            polygon: createPolygon(room),
        }));

        const filteredRooms = turfRooms.filter((roomA, i) => {
            // If any other room is strictly inside roomA, mark roomA for deletion
            return !turfRooms.some((roomB, j) => {
                if (i === j) return false;
                // Ensure we’re not comparing same polygon
                const areaA = turf.area(roomA.polygon);
                const areaB = turf.area(roomB.polygon);

                if (areaB >= areaA) return false;

                return turf.booleanWithin(turf.centerOfMass(roomB.polygon), roomA.polygon);
            });
        });

        return filteredRooms.map(room => room.points);
    };

    // Check if a cycle already exists in our list (considering different orders)
    const cycleExists = (allCycles: string[][], newCycle: string[]) => {
        const newSet = new Set(newCycle);

        return allCycles.some(existingCycle => {
            if (existingCycle.length !== newCycle.length) return false;
            const existingSet = new Set(existingCycle);
            return setsEqual(newSet, existingSet);
        });
    };

    // Check if two sets are equal
    const setsEqual = <T,>(a: Set<T>, b: Set<T>) => {
        if (a.size !== b.size) return false;
        for (const item of a) if (!b.has(item)) return false;
        return true;
    };

    // Remove duplicate rooms (same set of points in different orders)
    const removeDuplicateRooms = (rooms: Point[][]) => {
        const uniqueRooms: Point[][] = [];
        const roomHashes = new Set<string>();

        for (const room of rooms) {
            // Create a consistent hash for the room regardless of point order
            const hash = room
                .map(p => p.pointUuid)
                .sort()
                .join('-');

            if (!roomHashes.has(hash)) {
                roomHashes.add(hash);
                uniqueRooms.push(room);
            }
        }

        return uniqueRooms;
    };

    // Check if a room is valid (closed, non-self-intersecting polygon)
    const isValidRoom = (points: Point[]): boolean => {
        // Must have at least 4 points (first and last are same)
        if (points.length < 4) return false;

        // Must be a closed loop
        if (points[0].pointUuid !== points[points.length - 1].pointUuid) {
            return false;
        }

        try {
            const coordinates = points.map(p => [p.position[0], p.position[2]]);
            const polygon = turf.polygon([coordinates]);
            return turf.booleanValid(polygon);
        } catch (e) {
            return false;
        }
    };

    // Rest of the implementation remains the same...
    const rooms = useMemo(() => findRooms(), [walls]);

    function createPolygon(points: Point[]) {
        const coordinates = points.map(p => [p.position[0], p.position[2]]);
        return turf.polygon([coordinates]);
    };


    // Get all walls that form part of any room
    const getRoomWalls = (): Wall[] => {
        const roomWalls = new Set<Wall>();

        rooms.forEach(room => {
            for (let i = 0; i < room.length - 1; i++) {
                const p1 = room[i];
                const p2 = room[i + 1];

                // Find the wall that connects these two points
                const wall = walls.find(w =>
                    (w.points[0].pointUuid === p1.pointUuid && w.points[1].pointUuid === p2.pointUuid) ||
                    (w.points[0].pointUuid === p2.pointUuid && w.points[1].pointUuid === p1.pointUuid)
                );

                if (wall) roomWalls.add(wall);
            }
        });

        return Array.from(roomWalls);
    };

    // Determine wall orientation relative to room
    const getWallOrientation = (wall: Wall) => {
        const roomWalls = getRoomWalls();
        const isRoomWall = roomWalls.includes(wall);

        if (!isRoomWall) {
            return {
                isRoomWall: false,
                isOutsideFacing: false
            };
        }

        // Find which room this wall belongs to
        const containingRoom = rooms.find(room => {
            for (let i = 0; i < room.length - 1; i++) {
                const p1 = room[i];
                const p2 = room[i + 1];
                if ((wall.points[0].pointUuid === p1.pointUuid && wall.points[1].pointUuid === p2.pointUuid) ||
                    (wall.points[0].pointUuid === p2.pointUuid && wall.points[1].pointUuid === p1.pointUuid)) {
                    return true;
                }
            }
            return false;
        });

        if (!containingRoom) {
            return {
                isRoomWall: false,
                isOutsideFacing: false
            };
        }

        // Calculate the normal vector to determine outside direction
        const roomPoints = containingRoom.map(p => new THREE.Vector3(p.position[0], 0, p.position[2]));
        const centroid = new THREE.Vector3();
        roomPoints.forEach(p => centroid.add(p));
        centroid.divideScalar(roomPoints.length);

        const [p1, p2] = wall.points;
        const wallVector = new THREE.Vector3(
            p2.position[0] - p1.position[0],
            0,
            p2.position[2] - p1.position[2]
        );
        const normal = new THREE.Vector3(-wallVector.z, 0, wallVector.x).normalize();

        // Check if normal points away from centroid (outside)
        const testPoint = new THREE.Vector3(
            (p1.position[0] + p2.position[0]) / 2 + normal.x,
            0,
            (p1.position[2] + p2.position[2]) / 2 + normal.z
        );

        const pointInside = turf.booleanPointInPolygon(
            turf.point([testPoint.x, testPoint.z]),
            createPolygon(containingRoom)
        );

        // Determine if the wall is in the same order as the room's edge
        let isSameOrder = false;
        for (let i = 0; i < containingRoom.length - 1; i++) {
            const roomP1 = containingRoom[i];
            const roomP2 = containingRoom[i + 1];
            if (p1.pointUuid === roomP1.pointUuid && p2.pointUuid === roomP2.pointUuid) {
                isSameOrder = true;
                break;
            }
        }

        return {
            isRoomWall: true,
            isOutsideFacing: isSameOrder ? !pointInside : pointInside
        };
    };

    // Rest of the functions remain the same...
    const getWallType = (wall: Wall): 'room' | 'segment' => {
        return getWallOrientation(wall).isRoomWall ? 'room' : 'segment';
    };

    const isRoomWall = (wall: Wall): boolean => {
        return getWallOrientation(wall).isRoomWall;
    };

    const isSegmentWall = (wall: Wall): boolean => {
        return !getWallOrientation(wall).isRoomWall;
    };

    const getWallMaterialSide = (wall: Wall): { front: 'inside' | 'outside', back: 'inside' | 'outside' } => {
        const orientation = getWallOrientation(wall);

        if (!orientation.isRoomWall) {
            return { front: 'inside', back: 'outside' };
        }

        return orientation.isOutsideFacing
            ? { front: 'outside', back: 'inside' }
            : { front: 'inside', back: 'outside' };
    };

    return {
        rooms,
        getWallType,
        isRoomWall,
        isSegmentWall,
        getWallMaterialSide,
        findRooms,
    };
}